package io.tswf.groovy.bettergroovy.asm.util

import groovy.transform.CompileStatic
import org.objectweb.asm.Type
import org.objectweb.asm.tree.InsnList
import org.objectweb.asm.tree.InsnNode
import org.objectweb.asm.tree.JumpInsnNode
import org.objectweb.asm.tree.LabelNode
import org.objectweb.asm.tree.LdcInsnNode
import org.objectweb.asm.tree.MethodInsnNode
import org.objectweb.asm.tree.TypeInsnNode

import static org.objectweb.asm.Opcodes.*

@Singleton(lazy = true)
@CompileStatic
class BoxingUtil {
    Type unboxNullExceptionType = NullPointerException.asAsmType()

    Map<String, String> boxingMap = new HashMap<String, String>().tap {
        put(boolean.asmDescriptor, Boolean.asmDescriptor)
        put(char.asmDescriptor, Character.asmDescriptor)
        put(byte.asmDescriptor, Byte.asmDescriptor)
        put(short.asmDescriptor, Short.asmDescriptor)
        put(int.asmDescriptor, Integer.asmDescriptor)
        put(long.asmDescriptor, Long.asmDescriptor)
        put(float.asmDescriptor, Float.asmDescriptor)
        put(double.asmDescriptor, Double.asmDescriptor)
    }

    Map<String, String> unboxingMap = boxingMap.flip()

    String getBoxedDescriptorOrSelf(String unboxedDescriptor) {
        return boxingMap[unboxedDescriptor] ?: unboxedDescriptor
    }

    String getUnboxedDescriptorOrSelf(String boxedDescriptor) {
        return unboxingMap[boxedDescriptor] ?: boxedDescriptor
    }

    String getBoxedDescriptor(String unboxedDescriptor) {
        return boxingMap[unboxedDescriptor].ensureNotNull { new NoSuchElementException("There is no boxing for '${unboxedDescriptor}'! Here existing boxings: ${boxingMap}") }
    }

    String getUnboxedDescriptor(String boxedDescriptor) {
        return unboxingMap[boxedDescriptor].ensureNotNull { new NoSuchElementException("There is no unboxing for '${boxedDescriptor}'! Here existing unboxings: ${unboxingMap}") }
    }

    InsnList applyBoxing(String unboxedDescriptor, InsnList dest = new InsnList()) {
        def boxedType = Type.getType(getBoxedDescriptor(unboxedDescriptor))
        dest.add(new MethodInsnNode(INVOKESTATIC, boxedType.internalName, "valueOf", "(${unboxedDescriptor})${boxedType.descriptor}", false))

        return dest
    }

    InsnList applyUnboxing(String boxedDescriptor, InsnList dest = new InsnList()) {
        def unboxedType = Type.getType(getUnboxedDescriptor(boxedDescriptor))
        def boxedType = Type.getType(boxedDescriptor)

        def notNullLabel = new LabelNode()

        dest << new JumpInsnNode(IFNE, notNullLabel)

        // Если пытаемся заанбоксить null, то это исключение с понятным текстом, а не магическая ошибка
        dest << new TypeInsnNode(NEW, NullPointerException.asmInternalName)
        dest << new InsnNode(DUP)
        dest << new LdcInsnNode("Can't unbox null ${unboxedType.className} to ${boxedType.className}! Privitives can't be null!")
        dest << new MethodInsnNode(INVOKEVIRTUAL, unboxNullExceptionType.internalName, "<init>", "(${String.asmDescriptor})V")
        dest << new InsnNode(ATHROW)

        dest << notNullLabel
        dest.add(new MethodInsnNode(INVOKEVIRTUAL, boxedType.internalName, "${unboxedType.className}Value", "()${unboxedType.descriptor}", false))

        return dest
    }
}
